(** 

This file contains a representation of finite maps 
that may be used for representing a store. It also
provides lemmas and tactics for reasoning about
operations on the store (read, write, union). 

Author: Arthur Charguéraud.
License: MIT.

*)

Set Implicit Arguments.
Require Import LibCore.


(************************************************************)

Tactic Notation "cases" constr(E) :=  (* For TLC *)
  let H := fresh "Eq" in cases E as H.


(********************************************************************)
(* * States *)


(*------------------------------------------------------------------*)
(** Existence of fresh locations *)

(** These lemmas are useful to prove:
    [forall h v, exists l, state_disjoint (state_single l v) h]. *)

Definition loc_fresh_gen (L : list nat) :=
  (1 + fold_right plus O L)%nat.

Lemma loc_fresh_ind : forall l L,
  Mem l L -> (l < loc_fresh_gen L)%nat.
Proof using.
  intros l L. unfold loc_fresh_gen.
  induction L; introv M; inverts M; rew_list. 
  { math. }
  { forwards~: IHL. math. }
Qed.

Definition loc_fresh_property (A:Type) := 
  forall (L : list A), exists l, ~ Mem l L.

Lemma loc_fresh_nat : loc_fresh_property nat.
Proof using.
  intros L. exists (loc_fresh_gen L). intros M.
  lets: loc_fresh_ind M. math.
Qed.

Hint Resolve loc_fresh_nat.


(*------------------------------------------------------------------*)
(* ** Representation of partial functions *)

(** Type of partial functions from A to B *)

Definition pfun (A B : Type) :=
  A -> option B.

(** Finite domain of a partial function *)

Definition pfun_finite (A B : Type) (f : pfun A B) :=
  exists (L : list A), forall (x:A), f x <> None -> Mem x L.

(** Disjointness of domain of two partial functions *)

Definition pfun_disjoint (A B : Type) (f1 f2 : pfun A B) :=
  forall (x:A), f1 x = None \/ f2 x = None.

(** Disjoint union of two partial functions *)

Definition pfun_union (A B : Type) (f1 f2 : pfun A B) :=
  fun (x:A) => match f1 x with
           | Some y => Some y
           | None => f2 x
           end.

Definition pfun_agree (A B : Type) (f1 f2 : pfun A B) :=
  forall x v1 v2, 
  f1 x = Some v1 ->
  f2 x = Some v2 ->
  v1 = v2.


(*------------------------------------------------------------------*)
(** Representation of state *)

Section State. 
Variables loc : Type.
Variables val : Type.

(** Representation of heaps *)

Inductive state : Type := state_of {
  state_data :> pfun loc val;
  state_finite : pfun_finite state_data }.

Implicit Arguments state_of []. 

(** Disjoint states *)

Definition state_disjoint (h1 h2 : state) : Prop :=
  pfun_disjoint h1 h2.

Definition state_disjoint_3 h1 h2 h3 :=
  state_disjoint h1 h2 /\ state_disjoint h2 h3 /\ state_disjoint h1 h3.

(** Compatible states *)

Definition state_agree (g1 g2 : state) :=
  pfun_agree g1 g2.

(** Finiteness of union *)

Lemma pfun_union_finite : forall (A B : Type) (f1 f2 : pfun A B),
  pfun_finite f1 -> pfun_finite f2 -> pfun_finite (pfun_union f1 f2).
Proof using.
  introv [L1 F1] [L2 F2]. exists (L1 ++ L2). introv M.
  specializes F1 x. specializes F2 x. unfold pfun_union in M. 
  apply Mem_app_or. destruct~ (f1 x).
Qed.

(** Union of states *)

Program Definition state_union (h1 h2 : state) : state :=
  state_of (pfun_union h1 h2) _.
Next Obligation. destruct h1. destruct h2. apply~ pfun_union_finite. Qed.

(** Empty state *)

Program Definition state_empty : state :=
  state_of (fun l => None) _.
Next Obligation. exists (@nil loc). auto_false. Qed.

(** Singleton state *)

Program Definition state_single (l:loc) (v:val) : state := 
  state_of (fun l' => If l = l' then Some v else None) _.
Next Obligation.
  exists (l::nil). intros. case_if. subst~.
Qed.

(** Update of a state *)

Definition state_update (h:state) (l:loc) (v:val) := 
  state_union (state_single l v) h. 
  (* Note: the union operation first reads in the first argument. *)

(** Inhabited type [state] *)

Global Instance state_inhab : Inhab state.
Proof using. intros. applys prove_Inhab state_empty. Qed.

Notation "\# h1 h2" := (state_disjoint h1 h2)
  (at level 40, h1 at level 0, h2 at level 0, no associativity) : state_scope.

Notation "\# h1 h2 h3" := (state_disjoint_3 h1 h2 h3)
  (at level 40, h1 at level 0, h2 at level 0, h3 at level 0, no associativity)
  : state_scope.

Notation "h1 \+ h2" := (state_union h1 h2)
   (at level 51, right associativity) : state_scope.

Open Scope state_scope.


(************************************************************)
(* * Lemmas *)

(*------------------------------------------------------------------*)
(* ** Equality of states *)

(** Symmetry of disjointness *)

Lemma pfun_disjoint_sym : forall (A B : Type),
  sym (@pfun_disjoint A B).
Proof using.
  introv H. unfolds pfun_disjoint. intros z. specializes H z. intuition.
Qed.

(** Commutativity of disjoint union *)

Lemma pfun_union_comm : forall (A B : Type) (f1 f2 : pfun A B),
  pfun_disjoint f1 f2 -> 
  pfun_union f1 f2 = pfun_union f2 f1.
Proof using. 
  introv H. extens. intros x. unfolds pfun_disjoint, pfun_union.
  specializes H x. cases (f1 x); cases (f2 x); auto. destruct H; false. 
Qed.

(** Proving states equals *)

Lemma state_eq : forall (f1 f2: loc -> option val) F1 F2,
  (forall x, f1 x = f2 x) ->
  state_of f1 F1 = state_of f2 F2.
Proof using.
  introv H. asserts: (f1 = f2). extens~. subst. 
  fequals. (* note: involves proof irrelevance *)
Qed.


(*------------------------------------------------------------------*)
(* ** Properties on states *)

(** Disjointness *)

Lemma state_disjoint_sym : forall h1 h2,
  \# h1 h2 -> \# h2 h1.
Proof using. intros [f1 F1] [f2 F2]. apply pfun_disjoint_sym. Qed.

Lemma state_disjoint_comm : forall h1 h2,
  \# h1 h2 = \# h2 h1.
Proof using. lets: state_disjoint_sym. extens*. Qed. 

Lemma state_disjoint_empty_l : forall h,
  \# state_empty h.
Proof using. intros [f F] x. simple~. Qed.

Lemma state_disjoint_empty_r : forall h,
  \# h state_empty.
Proof using. intros [f F] x. simple~. Qed.

Hint Resolve state_disjoint_sym state_disjoint_empty_l state_disjoint_empty_r.

Lemma state_disjoint_union_eq_r : forall h1 h2 h3,
  \# h1 (h2 \+ h3) =
  (\# h1 h2 /\ \# h1 h3).
Proof using.
  intros [f1 F1] [f2 F2] [f3 F3].
  unfolds state_disjoint, state_union. simpls.
  unfolds pfun_disjoint, pfun_union. extens. iff M [M1 M2].
  split; intros x; specializes M x; 
   destruct (f2 x); intuition; tryfalse.
  intros x. specializes M1 x. specializes M2 x. 
   destruct (f2 x); intuition.
Qed.

Lemma state_disjoint_union_eq_l : forall h1 h2 h3,
  \# (h2 \+ h3) h1 =
  (\# h1 h2 /\ \# h1 h3).
Proof using.
  intros. rewrite state_disjoint_comm. 
  apply state_disjoint_union_eq_r.
Qed.

Lemma state_single_same_loc_disjoint : forall (l:loc) (v1 v2:val),
  \# (state_single l v1) (state_single l v2) -> False.
Proof using.
  introv D. specializes D l. simpls. case_if. destruct D; tryfalse.
Qed.

Lemma state_disjoint_3_unfold : forall h1 h2 h3,
  \# h1 h2 h3 = (\# h1 h2 /\ \# h2 h3 /\ \# h1 h3).
Proof using. auto. Qed.

Lemma state_disjoint_new : forall null (h:state) v,
  loc_fresh_property loc ->
  exists l, \# (state_single l v) h /\ l <> null.
Proof using.
  intros null (m&(L&M)) v HF.
  unfold state_disjoint, pfun_disjoint. simpl.
  lets (l&F): (HF (null::L)).
  exists l. split.
  { intros l'. case_if~. 
    { right. applys not_not_elim. intros H. applys F.
      constructor. applys~ M. } }
  { intro_subst. applys~ F. }
Qed.

Lemma state_disjoint_single_set : forall (h:state) l v1 v2,
  \# (state_single l v1) h ->
  \# (state_single l v2) h.
Proof using.
  introv M. unfolds state_disjoint, state_single, pfun_disjoint; simpls.  
  intros l'. specializes M l'. case_if~. destruct M; auto_false.
Qed.

(** Union *)

Lemma state_union_idempotent : forall h,
  h \+ h = h.
Proof using.
  intros [f F]. apply~ state_eq. simpl. intros x.
  unfold pfun_union. cases~ (f x).
Qed.

Lemma state_union_empty_l : forall h,
  state_empty \+ h = h.
Proof using. 
  intros [f F]. unfold state_union, pfun_union, state_empty. simpl.
  apply~ state_eq.
Qed.

Lemma state_union_empty_r : forall h,
  h \+ state_empty = h.
Proof using. 
  intros [f F]. unfold state_union, pfun_union, state_empty. simpl.
  apply state_eq. intros x. destruct~ (f x).
Qed.

Lemma state_union_eq_empty_inv_l : forall h1 h2,
  h1 \+ h2 = state_empty ->
  h1 = state_empty.
Proof using.
  intros (f1&F1) (f2&F2) M. inverts M as M.
  applys state_eq. intros l. 
  unfolds pfun_union. 
  lets: func_same_1 l M.
  cases (f1 l); auto_false.
Qed.

Lemma state_union_eq_empty_inv_r : forall h1 h2,
  h1 \+ h2 = state_empty ->
  h2 = state_empty.
Proof using.
  intros (f1&F1) (f2&F2) M. inverts M as M.
  applys state_eq. intros l. 
  unfolds pfun_union. 
  lets: func_same_1 l M.
  cases (f1 l); auto_false.
Qed.

Lemma state_union_comm_disjoint : forall h1 h2,
  \# h1 h2 ->
  h1 \+ h2 = h2 \+ h1.
Proof using.
  intros [f1 F1] [f2 F2] H. simpls. apply state_eq. simpl.
  intros. rewrite~ pfun_union_comm.
Qed.

Lemma state_union_comm_agree : forall h1 h2,
  state_agree h1 h2 ->
  h1 \+ h2 = h2 \+ h1.
Proof using.
  intros [f1 F1] [f2 F2] H. simpls. apply state_eq. simpl.
  intros l. specializes H l. unfolds pfun_union. simpls.
  cases (f1 l); cases (f2 l); auto. fequals. applys* H.
Qed.

Lemma state_union_assoc : forall h1 h2 h3,
  (h1 \+ h2) \+ h3 = h1 \+ (h2 \+ h3).
Proof using.
  intros [f1 F1] [f2 F2] [f3 F3]. unfolds state_union. simpls.
  apply state_eq. intros x. unfold pfun_union. destruct~ (f1 x).
Qed.

Lemma pfun_eq_forward : forall (h1 h2 : state),
  h1 = h2 ->
  forall x, state_data h1 x = state_data h2 x.
Proof using. intros. fequals. Qed.

(*
Lemma state_union_eq_inv : forall h2 h1 h1' : state,
  \# h1 h2 ->
  state_agree h1' h2 ->
  h1 \+ h2 = h1' \+ h2 ->
  h1 = h1'.
Proof using.
  intros [f2 F2] [f1 F1] [f1' F1'] D D' E.
  applys state_eq. intros x. specializes D x. specializes D' x.
  lets E': pfun_eq_forward (rm E) x. simpls.
  unfolds pfun_union. 
  cases (f1 x); cases (f2 x); try solve [cases (f1' x); destruct D; congruence ].
  destruct D; try false.
  rewrite H in E'. inverts E'.
  cases (f1' x); cases (f1 x);
    destruct D; try congruence.
  false.
    destruct D'; try congruence.
Qed.
*)

Lemma state_union_eq_inv : forall h2 h1 h1' : state,
  \# h1 h2 ->
  \# h1' h2 ->
  h1 \+ h2 = h1' \+ h2 ->
  h1 = h1'.
Proof using.
  intros [f2 F2] [f1' F1'] [f1 F1] D D' E.
  applys state_eq. intros x. specializes D x. specializes D' x.
  lets E': pfun_eq_forward (rm E) x. simpls.
  unfolds pfun_union. 
  cases (f1' x); cases (f1 x);
    destruct D; try congruence;
    destruct D'; try congruence.
Qed.

(** Compatibility *)

Lemma state_agree_refl : forall h,
  state_agree h h.
Proof using.
  intros h. introv M1 M2. congruence.
Qed.

Lemma state_agree_sym : forall f1 f2,
  state_agree f1 f2 ->
  state_agree f2 f1.
Proof using.
  introv M. intros l v1 v2 E1 E2.
  specializes M l E1. 
Qed.

Lemma state_agree_disjoint : forall h1 h2,
  state_disjoint h1 h2 ->
  state_agree h1 h2.
Proof using. 
  introv HD. intros l v1 v2 M1 M2. destruct (HD l); false.
Qed.

Lemma state_agree_empty_l : forall h,
  state_agree state_empty h.
Proof using. intros h l v1 v2 E1 E2. simpls. false. Qed.

Lemma state_agree_empty_r : forall h,
  state_agree h state_empty.
Proof using.
  hint state_agree_sym, state_agree_empty_l. eauto.
Qed.

Lemma state_agree_union_l : forall f1 f2 f3,
  state_agree f1 f3 ->
  state_agree f2 f3 ->
  state_agree (f1 \+ f2) f3. 
Proof using.
  introv M1 M2. intros l v1 v2 E1 E2.
  specializes M1 l. specializes M2 l.
  simpls. unfolds pfun_union. cases (state_data f1 l).
  { inverts E1. applys* M1. }
  { applys* M2. }
Qed.

Lemma state_agree_union_r : forall f1 f2 f3,
  state_agree f1 f2 ->
  state_agree f1 f3 ->
  state_agree f1 (f2 \+ f3). 
Proof using.
  hint state_agree_sym, state_agree_union_l. eauto.
Qed.

Lemma state_agree_union_lr : forall f1 g1 f2 g2,
  state_agree g1 g2 ->
  \# f1 f2 (g1 \+ g2) ->
  state_agree (f1 \+ g1) (f2 \+ g2).
Proof using.
  introv M1 (M2a&M2b&M2c).
  rewrite state_disjoint_union_eq_r in *.
  applys state_agree_union_l; applys state_agree_union_r;
    try solve [ applys* state_agree_disjoint ].
  auto.
Qed.

Lemma state_agree_union_ll_inv : forall f1 f2 f3,
  state_agree (f1 \+ f2) f3 ->
  state_agree f1 f3.
Proof using.
  introv M. intros l v1 v2 E1 E2.
  specializes M l. simpls. unfolds pfun_union.
  rewrite E1 in M. applys* M.
Qed.

Lemma state_agree_union_rl_inv : forall f1 f2 f3,
  state_agree f1 (f2 \+ f3) ->
  state_agree f1 f2.
Proof using.
  hint state_agree_union_ll_inv, state_agree_sym. eauto.
Qed.

Lemma state_agree_union_lr_inv_agree_agree : forall f1 f2 f3,
  state_agree (f1 \+ f2) f3 ->
  state_agree f1 f2 ->
  state_agree f2 f3.
Proof using.
  introv M D. rewrite~ (@state_union_comm_agree f1 f2) in M.
  applys* state_agree_union_ll_inv.
Qed.

Lemma state_agree_union_rr_inv_agree : forall f1 f2 f3,
  state_agree f1 (f2 \+ f3) ->
  state_agree f2 f3 ->
  state_agree f1 f3.
Proof using.
  hint state_agree_union_lr_inv_agree_agree, state_agree_sym. eauto.
Qed.

Lemma state_agree_union_l_inv : forall f1 f2 f3,
  state_agree (f1 \+ f2) f3 ->
  state_agree f1 f2 ->
     state_agree f1 f3 
  /\ state_agree f2 f3.
Proof using.
  (* TODO: proofs redundant with others above *)
  introv M2 M1. split.
  { intros l v1 v2 E1 E2. 
    specializes M1 l v1 v2 E1. applys~ M2 l v1 v2.
    unfold state_union, pfun_union; simpl. rewrite~ E1. }
  { intros l v1 v2 E1 E2.
    specializes M1 l. specializes M2 l. 
    unfolds state_union, pfun_union; simpls.
    cases (state_data f1 l).
    { applys eq_trans v. symmetry. applys~ M1. applys~ M2. } 
    { auto. } }
Qed.

Lemma state_agree_union_r_inv : forall f1 f2 f3,
  state_agree f1 (f2 \+ f3) ->
  state_agree f2 f3 ->
     state_agree f1 f2 
  /\ state_agree f1 f3.
Proof using. 
  hint state_agree_sym.
  intros. forwards~ (M1&M2): state_agree_union_l_inv f2 f3 f1.
Qed.

(** Read and write operations *)

Lemma state_union_single_read : forall f1 f2 l v, 
  f1 = state_single l v ->
  state_data (f1 \+ f2) l = Some v.
Proof using.
  intros. subst. simpl. unfold pfun_union. case_if~.
Qed.

Lemma state_union_single_write : forall f1 f1' f2 l v v', 
  f1 = state_single l v ->
  f1' = state_single l v' ->
  (f1' \+ f2) = state_update (f1 \+ f2) l v'.
Proof using.
  intros. subst. unfold state_update.
  rewrite <- state_union_assoc. fequals. 
  applys state_eq. intros l'.
  unfolds pfun_union, state_single; simpl. case_if~.
Qed.

End State.

Implicit Arguments state_empty [[loc] [val]].
Implicit Arguments state_union_assoc [loc val].
Implicit Arguments state_union_comm_disjoint [loc val].
Implicit Arguments state_union_comm_agree [loc val].


(************************************************************)
(* * Notation *)

Notation "\# h1 h2" := (state_disjoint h1 h2)
  (at level 40, h1 at level 0, h2 at level 0, no associativity) : state_scope.

Notation "\# h1 h2 h3" := (state_disjoint_3 h1 h2 h3)
  (at level 40, h1 at level 0, h2 at level 0, h3 at level 0, no associativity)
  : state_scope.

Notation "h1 \+ h2" := (state_union h1 h2)
   (at level 51, right associativity) : state_scope.

Open Scope state_scope.


(************************************************************)
(* * Tactics *)

(*------------------------------------------------------------------*)
(* ** Proving disjointness results *)

(** [state_disjoint] proves goals of the form [\# h1 h2] and
    [\# h1 h2 h3] by expanding all hypotheses into binary forms
    [\# h1 h2] and then exploiting symmetry and disjointness
    with [state_empty]. *)

Hint Resolve state_disjoint_sym state_disjoint_empty_l state_disjoint_empty_r.

Hint Rewrite 
  state_disjoint_union_eq_l
  state_disjoint_union_eq_r
  state_disjoint_3_unfold : rew_disjoint.

Tactic Notation "rew_disjoint" :=
  autorewrite with rew_disjoint in *.
Tactic Notation "rew_disjoint" "*" :=
  rew_disjoint; auto_star.

Tactic Notation "state_disjoint" :=
  solve [ subst; rew_disjoint; jauto_set; auto ].

Tactic Notation "state_disjoint_if_needed" :=
  match goal with 
  | |- \# _ _ => state_disjoint
  | |- \# _ _ _ => state_disjoint
  end.

Lemma state_disjoint_demo : forall A B (h1 h2 h3 h4 h5:state A B),
  h1 = h2 \+ h3 ->
  \# h2 h3 ->
  \# h1 h4 h5 ->
  \# h3 h2 h5 /\ \# h4 h5.
Proof using.
  intros. dup 2.
  { subst. rew_disjoint. jauto_set. auto. auto. auto. auto. }
  { state_disjoint. }
Qed. 


(*------------------------------------------------------------------*)
(* ** Tactic [state_eq] for proving equality of states,
      and tactic [rew_state] to normalize state expressions. *)

Section StateEq.
Variables (A B : Type).
Implicit Types h : state A B.

(** [state_eq] proves equalities between unions of states.
    It attempts to discharge the disjointness side-conditions.
    Disclaimer: cancels heaps at depth up to 4, but no more. *)

Lemma state_union_eq_cancel_1 : forall h1 h2 h2',
  h2 = h2' ->
  h1 \+ h2 = h1 \+ h2'.
Proof using. intros. subst. auto. Qed.  

Lemma state_union_eq_cancel_1' : forall h1,
  h1 = h1.
Proof using. intros. auto. Qed.  

Lemma state_union_eq_cancel_2 : forall h1 h1' h2 h2',
  \# h1 h1' ->
  h2 = h1' \+ h2' ->
  h1 \+ h2 = h1' \+ h1 \+ h2'.
Proof using.
  intros. subst. rewrite <- state_union_assoc.
  rewrite (state_union_comm_disjoint h1 h1').
  rewrite~ state_union_assoc. auto.
Qed.  

Lemma state_union_eq_cancel_2' : forall h1 h1' h2,
  \# h1 h1' ->
  h2 = h1' ->
  h1 \+ h2 = h1' \+ h1.
Proof using.
  intros. subst. apply~ state_union_comm_disjoint.
Qed.  

Lemma state_union_eq_cancel_3 : forall h1 h1' h2 h2' h3',
  \# h1 (h1' \+ h2') ->
  h2 = h1' \+ h2' \+ h3' ->
  h1 \+ h2 = h1' \+ h2' \+ h1 \+ h3'.
Proof using.
  intros. subst.
  rewrite <- (state_union_assoc h1' h2' h3').
  rewrite <- (state_union_assoc h1' h2' (h1 \+ h3')).
  apply~ state_union_eq_cancel_2.
Qed.

Lemma state_union_eq_cancel_3' : forall h1 h1' h2 h2',
  \# h1 (h1' \+ h2') ->
  h2 = h1' \+ h2' ->
  h1 \+ h2 = h1' \+ h2' \+ h1.
Proof using.
  intros. subst.
  rewrite <- (state_union_assoc h1' h2' h1).
  apply~ state_union_eq_cancel_2'.
Qed.  

Lemma state_union_eq_cancel_4 : forall h1 h1' h2 h2' h3' h4',
  \# h1 ((h1' \+ h2') \+ h3') ->
  h2 = h1' \+ h2' \+ h3' \+ h4' ->
  h1 \+ h2 = h1' \+ h2' \+ h3' \+ h1 \+ h4'.
Proof using.
  intros. subst.
  rewrite <- (state_union_assoc h1' h2' (h3' \+ h4')).
  rewrite <- (state_union_assoc h1' h2' (h3' \+ h1 \+ h4')).
  apply~ state_union_eq_cancel_3.
Qed.  

Lemma state_union_eq_cancel_4' : forall h1 h1' h2 h2' h3',
  \# h1 (h1' \+ h2' \+ h3') ->
  h2 = h1' \+ h2' \+ h3' ->
  h1 \+ h2 = h1' \+ h2' \+ h3' \+ h1.
Proof using.
  intros. subst.
  rewrite <- (state_union_assoc h2' h3' h1).
  apply~ state_union_eq_cancel_3'.
Qed.  

End StateEq.

Hint Rewrite 
  state_union_assoc
  state_union_empty_l
  state_union_empty_r : rew_state.

Tactic Notation "rew_state" :=
  autorewrite with rew_state in *.

Tactic Notation "rew_state" "~" :=
  rew_state; auto_tilde.

Tactic Notation "rew_state" "*" :=
  rew_state; auto_star.

Ltac state_eq_step tt :=
  match goal with | |- ?G => match G with
  | ?h1 = ?h1 => apply state_union_eq_cancel_1'
  | ?h1 \+ ?h2 = ?h1 \+ ?h2' => apply state_union_eq_cancel_1
  | ?h1 \+ ?h2 = ?h1' \+ ?h1 => apply state_union_eq_cancel_2'
  | ?h1 \+ ?h2 = ?h1' \+ ?h1 \+ ?h2' => apply state_union_eq_cancel_2
  | ?h1 \+ ?h2 = ?h1' \+ ?h2' \+ ?h1 => apply state_union_eq_cancel_3'
  | ?h1 \+ ?h2 = ?h1' \+ ?h2' \+ ?h1 \+ ?h3' => apply state_union_eq_cancel_3
  | ?h1 \+ ?h2 = ?h1' \+ ?h2' \+ ?h3' \+ ?h1 => apply state_union_eq_cancel_4'
  | ?h1 \+ ?h2 = ?h1' \+ ?h2' \+ ?h3' \+ ?h1 \+ ?h4' => apply state_union_eq_cancel_4
  end end.

Tactic Notation "state_eq" :=
  subst; 
  rew_state; 
  repeat (state_eq_step tt);
  try state_disjoint_if_needed.

Tactic Notation "state_eq" "~" :=
  state_eq; auto_tilde.

Tactic Notation "state_eq" "*" :=
  state_eq; auto_star.

Lemma state_eq_demo : forall A B (h1 h2 h3 h4 h5:state A B),
  \# h1 h2 h3 ->
  \# (h1 \+ h2 \+ h3) h4 h5 ->
  h1 = h2 \+ h3 ->
  h4 \+ h1 \+ h5 = h2 \+ h5 \+ h4 \+ h3.
Proof using.
  intros. dup 2.
  { subst. rew_state.
    state_eq_step tt. state_disjoint.
    state_eq_step tt.
    state_eq_step tt. state_disjoint. auto. }
  { state_eq. }
Qed. 

(*------------------------------------------------------------------*)
(* ** Tactic [state_red] for proving [red] goals modulo 
      equalities between states *)

(** [state_red] proves a goal of the form [red h1 t h2 v]
    using an hypothesis of the shape [red h1' t h2' v],
    generating [h1 = h1'] and [h2 = h2'] as subgoals, and
    attempting to solve them using the tactic [state_red]. *)

Ltac state_red_base tt := fail.
(* Example:
Ltac state_red_base tt :=
  match goal with H: red _ ?t _ _ |- red _ ?t _ _ =>
    applys_eq H 2 4; try state_eq end.
*)

Tactic Notation "state_red" := 
  state_red_base tt.

Tactic Notation "state_red" "~" := 
  state_red; auto_tilde.

Tactic Notation "state_red" "*" := 
  state_red; auto_star.
